#
# denon.attr - Denon/Marantz AV Receiver control - value attributes.
#
# Copyright 2012-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import core
import values

SEL_MAIN = 0
SEL_ZONE = 1


#
# Append an optional string to a pair of strings.
# Leave None values alone. Take a single string to mean a pair (value, None).
#
def normalize(source, append=''):
	if source is None:
		return None
	if isinstance(source, basestring):
		source = (source, None)
	return map(lambda s: None if s is None else s + append, source)


#
# An attribute descriptor for some AVR value
#
class Attribute(object):
	""" Base class of all AVR attributes.

		This base class is a full implementation for arbitrary string values.
		Subclasses may specialize for different values and various conversion
		behaviors.

		Type= is the type of value stored for the attribute. Convert= is a conversion
		class that provides for translation from wire values to stored values and back.
		The default is to just keep the string value as defined by Denon.

		All of cmd, scan, and query may be a single value for global Attributes,
		or a pair for (main, zone), where the zone value is automatically prefixed
		with Zn.
		If cmd is None, this is a read-only attribute. Otherwise, these are the
		commands to request value changes.
		If scan is None, it defaults to cmd + some word (\S+). This default requires
		cmd to be given. (A non-read, non-write Attribute is useless.)
		If query is False, there is no (working) way to query the value of the attribute.
		Otherwise, query defaults to "cmd?".
		The pri(ority) determines how soon the auto-discovery algorithm asks for
		the Attribute's value. Attributes with higher priorities are queried earlier.
		The count is the anticipated number of event replies triggered by a query.
		It is used to pace queries to avoid overflowing buffers.
		Events are usually filtered so only genuine changes are reported upstream.
		A raw attribute will report every received events even if it's a repeat of
		prior values.
		Edit_send is a set of filters that can edit an outgoing command (string).
		Likewise, edit_update contains filters that edit incoming event (regex match) data.
	"""
	def __init__(self,
		type=str,				# stored type
		convert=None,			# convert to/from wire format
		cmd=None,				# change command: main or (main, zone)
		scan=None,				# event recognizer: main or (main, zone)
		query=None,				# query command: main or (main, zone)
		default=None,			# initial value
		raw=False,				# deliver all events (do not de-bounce)
		edit_send=None,			# command edit filters
		edit_update=None,		# event reception filters
		pri=0,					# discovery priority (higher=more urgent)
		count=1					# expected number of responses (delay management)
		):
		self.type = type
		self.convert = convert if convert else self
		self.default = default
		self.raw = raw
		self.edit_send = edit_send or ()
		self.edit_update = edit_update or ()
		self.pri = pri
		self.count = count

		self.cmd = normalize(cmd)
		if scan is None and cmd is not None:
			if type == int or type == float:
				re = r'\d'
			else:
				re = r'[^\r]'
			self.scan = normalize(self.cmd, r'(%s+)' % re)
		elif scan == False:
			self.scan = None
		else:
			self.scan = normalize(scan)
		self.query = None if query == False else normalize(self.cmd, "?") if query == None else normalize(query)

	@property
	def writable(self):
		return self.cmd is not None

	@property
	def queryable(self):
		return self.query is not None

	def __get__(self, obj, cls):
		""" Retrieve last reported value. """
		return obj._attr_values.get(self.name)

	def __set__(self, obj, value):
		""" Request a change in value. Note that the actual update is asynchronous upon success. """
		cmd = self.cmd[obj.zone_sel]
		if cmd is None:
			raise AttributeError('"%s" is read-only' % self.name)
		cmd += self.convert.to_string(value, obj.zone_sel == SEL_MAIN)
		for filter in self.edit_send:
			cmd = filter(obj, cmd)
		obj.command(cmd)

	def update(self, obj, arg):
		""" Update value from a device report. This default uses a single value. """
		obj._attr_values[self.name] = self.convert.to_value(arg)

	# default converter behavior (built-in)
	def to_value(self, s):
		return s.rstrip(' \t\f\r\n\0')
	def to_string(self, s, main):
		return str(s)

	def __repr__(self):
		return "<attr:%s>" % self.name


#
# A Metaclass that collects all our descriptors
#
class AttributedClass(type):
    """ A metaclass to collect our attribute Descriptors from class definitions.

        Specifically, collect any descriptors of type Attribute from this
        or any base class, except ones whose name starts with _,
        and put them into cls.attributes.
    """
    @staticmethod
    def _collect_descriptors(cls):
        descmap = { }
        # pick up descriptors from base classes
        for parent in cls.__bases__:        # inherit descriptors, if any
            if 'attributes' in parent.__dict__:
                descmap.update(parent.attributes)
        # pick up our own
        for (name, attr) in [(k, v) for (k, v) in cls.__dict__.items()
            if k[0] != '_' and isinstance(v, Attribute)]:
            attr.name = name
            descmap[name] = attr
        return descmap

    def __init__(cls, name, bases, content):
		type.__init__(cls, name, bases, content)
		cls.attributes = AttributedClass._collect_descriptors(cls)


#
# Base class for all AVR device classes
#
class Device(object):

	__metaclass__ = AttributedClass

	def __init__(self):
		self._attr_values = { }
		for attr in [attr for attr in type(self).attributes.values() if attr.default is not None]:
			self._attr_values[attr.name] = attr.default

	def has_attribute(self, attr):
		""" Does this device have this attribute. Takes either name or Attribute object. """
		if isinstance(attr, Attribute):
			attr = attr.name
		return attr in self._attr_values

	def missing_attributes(self):
		""" A list of all attributes of this device that have no values. """
		return [attr for attr in self.attributes.values() if not self.has_attribute(attr)]

	def set_attribute(self, name, value):
		""" setattr(self, name, value) but with intelligent type conversion for value. """
		attr = self.attributes[name]	# [KeyError]
		if attr.type == bool and isinstance(value, basestring):
			value = value.lower() in ['yes', 'true', 'on']
		elif attr.type in [int, float]:
			value = attr.type(value)
		setattr(self, name, value)


#
# An attribute describing a sound volume (in dB-relative form)
#
class VolumeAttribute(Attribute):
	def __init__(self, zero=core.CHAN_ZERO, fine=(False, False), **kwargs):
		Attribute.__init__(self, type=float, convert=values.DenonVolume(zero=zero, fine=fine), **kwargs)


#
# An attribute that is either on (True) or off (False)
#
class BooleanAttribute(Attribute):
	def __init__(self, cmd=None, scan=None, off='OFF', **kwargs):
		if scan is None:
			scan = "%s(ON|%s)" % (cmd, off)
		Attribute.__init__(self, type=bool, cmd=cmd, scan=scan,
			convert=values.Toggle(off=off), **kwargs)


#
# An attribute whose string value is restricted to a list of allowed values
#
class EnumAttribute(Attribute):
	def __init__(self, choices, type=str, cmd=None, scan=None, **kwargs):
		if scan is None and cmd is not None:
			scan = normalize(cmd, "(%s)" % "|".join(map(str, choices)))
		Attribute.__init__(self, type=type, cmd=cmd, scan=scan, **kwargs)
		self.choices = choices

	def __set__(self, obj, value):
		if value not in self.choices:
			raise ValueError('%s must be one of %s, not "%s"' % (
				self.name, '/'.join(self.choices), value))
		return Attribute.__set__(self, obj, value)


#
# An attribute whose value is a dictionary.
# This currently does not support assignment (__set__).
#
class DictAttribute(Attribute):
	def __init__(self, key_type=str, key_convert=None, **kwargs):
		Attribute.__init__(self, type=dict, **kwargs)
		self.key_type = key_type
		self.key_convert=key_convert or values.Converter()

	def update(self, obj, key, arg):
		if self.name not in obj._attr_values:
			obj._attr_values[self.name] = { }
		key = self.key_convert.to_value(self.key_type(key))
		obj._attr_values[self.name][key] = self.convert.to_value(arg)
