#
# denon.values - Denon/Marantz AV Receiver control control value handling
#
# Copyright 2012-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import math

from core import VOL_ZERO, VOL_DEFEAT, VOL_LOWEST
from core import VOLUME_DEFEAT
from core import PRESET_ROW


#
# The base class of all value managers
#
class Converter(object):
	""" A conversion object between Denon rep and internal value.

		to_value converts from Denon event string to internal value.
		to_string converts the other way (internal value to Denon command argument).

		This default simply uses the Denon string values as given.
	"""
	def to_value(self, s):
		return s

	def to_string(self, v, main):
		return v


#
# Long-form string values
#
class StringConverter(Converter):
	""" A Converter that deals with Denon line-strings. """

	def to_value(self, s):
		if not s or s[0] == '\0':
			return ''
		else:
			return s.rstrip(' \t\f\r\n\0')


#
# Boolean toggle values
#
class Toggle(Converter):
	""" A Converter that handles (only) two values - a toggle. """

	def __init__(self, on='ON', off='OFF'):
		self.on = on
		self.off = off

	def to_value(self, s):
		if s == self.on:
			return True
		elif s == self.off:
			return False
		else:
			raise ValueError(s)

	def to_string(self, v, main):
		return self.on if v else self.off


#
# A tuner frequency.
# AM (kHz) and FM (MHz) are overloaded since they're distinct ranges.
#
class Frequency(Converter):

	def to_value(self, code):
		return int(code) / 100.0

	def to_string(self, freq, main):
		return "%06d" % round(100.0 * freq)


#
# Audio volume values
#
class DenonVolume(Converter):
	""" A Converter that handles Denon volume indications.

		The encoding is idiosyncratic. It essentially maps a range of
		two-and-a-half digit precision [0..100) to, effectively, an absolute
		range of [-81..+19) dB where -81 is effectively "off".
	"""
	def __init__(self, zero=VOL_ZERO, fine=False):
		self._zero = zero
		self._fine = fine

	def to_value(self, code):
		""" Convert Denon volume code (nn or nn5) to dbm volume level. """
		if code == VOL_DEFEAT:
			return VOLUME_DEFEAT
		elif code == VOL_LOWEST:	# special (weird) case
			return -self._zero - 0.5
		else:
			return int(code[0:2]) + 0.1 * (int(code[2]) if len(code) > 2 else 0) - self._zero

	def to_string(self, volume, main):
		if volume <= VOLUME_DEFEAT:	# below minimum threshold
			return VOL_DEFEAT
		elif volume < -self._zero:
			return VOL_LOWEST
		else:
			(frac, level) = math.modf(volume + self._zero)
			base = '%02.2d' % level
			return (base + '5') if self._fine and main and frac >= 0.5 else base


#
# Internet preset codes (00-55 <=> A1-G8)
#
class InternetPreset(Converter):

	def to_string(self, name, main):
		if len(name) == 2 and name[0] in 'ABCDEFG' and name[1] in '12345678':
			return "%02d" % ((ord(name[0]) - ord('A')) * PRESET_ROW + ord(name[1]) - ord('1'))
		else:
			raise ValueError(name)

	def to_value(self, code):
		icode = int(code)
		return "%c%d" % (chr(ord('A') + icode / PRESET_ROW), icode % PRESET_ROW + 1)
